"""Accelerator for numpy format handler operations"""
#cython: language_level=3
from ctypes import c_void_p
import numpy as np
cimport numpy as np
from OpenGL_accelerate.formathandler cimport FormatHandler
import traceback, weakref
from OpenGL.error import CopyError
from OpenGL._bytes import bytes,unicode

cdef extern from "Python.h":
    cdef void Py_INCREF( object )

cdef extern from "numpy/arrayobject.h":
    cdef np.ndarray PyArray_FromArray( np.ndarray, np.dtype, int )
    cdef np.ndarray PyArray_ContiguousFromAny( object op, int, int, int max_depth)
    cdef int PyArray_Check( object )
    cdef int PyArray_CheckScalar( object )
    int NPY_ARRAY_CARRAY
    int NPY_ARRAY_FORCECAST
    int PyArray_ISCARRAY( np.ndarray instance )
    int PyArray_ISCARRAY_RO( np.ndarray instance )
    cdef np.ndarray PyArray_Zeros(int nd, np.npy_intp* dims, np.dtype, int fortran)
    cdef np.ndarray PyArray_EnsureArray(object)
    cdef int PyArray_FillWithScalar(np.ndarray, object)
    cdef void import_array()
    cdef void* PyArray_DATA( np.ndarray )
    cdef int PyArray_NDIM( np.ndarray )
    cdef int *PyArray_DIMS( np.ndarray )
    cdef int PyArray_DIM( np.ndarray, int dim )
    cdef np.dtype PyArray_DESCR( np.ndarray )
    cdef np.npy_intp PyArray_SIZE( np.ndarray )

cdef np.dtype array_descr( np.ndarray array ):
    """Wrap PyArray_DESCR and incref to deal with the "borrowed" reference"""
    cdef np.dtype desc = PyArray_DESCR( array )
    Py_INCREF(<object> desc)
    return desc

cdef class NumpyHandler(FormatHandler):
    cdef public dict array_to_gl_constant
    cdef public dict gl_constant_to_array
    isOutput = True
    HANDLED_TYPES = (
        np.ndarray,
        np.bool_,
        np.intc,
        np.uintc,
        np.int8,
        np.uint8,
        np.int16,
        np.uint16,
        np.int32,
        np.uint32,
        np.int64,
        np.uint64,
        np.int64,
        np.uint64,
        np.float16,
        np.float32,
        np.float64,
        np.complex64,
        np.complex128,
        np.bytes_,
        np.str_,
        np.void,
        np.datetime64,
        np.timedelta64,
    )
    if hasattr(np,'float128'):
        HANDLED_TYPES += (np.float128,)
    if hasattr(np,'complex256'):
        HANDLED_TYPES += (np.complex256,)
    
    def __init__( self, ERROR_ON_COPY=None, a_to_gl=None, gl_to_a=None ):
        if ERROR_ON_COPY is None:
            from OpenGL import _configflags
            ERROR_ON_COPY = _configflags.ERROR_ON_COPY
        if a_to_gl is None:
            from OpenGL.arrays.numpymodule import ARRAY_TO_GL_TYPE_MAPPING
            a_to_gl = ARRAY_TO_GL_TYPE_MAPPING
        if gl_to_a is None:
            from OpenGL.arrays.numpymodule import GL_TYPE_TO_ARRAY_MAPPING
            gl_to_a = GL_TYPE_TO_ARRAY_MAPPING
        self.ERROR_ON_COPY = ERROR_ON_COPY
        self.array_to_gl_constant = a_to_gl
        self.gl_constant_to_array = gl_to_a
    
    cdef np.ndarray c_check_array( self, object instance ):
        if not PyArray_Check( instance ):
            raise TypeError(
                """Numpy format handler passed a non-numpy-array object %s (of type %s)"""%( instance, type(instance) ),
            )
        return <np.ndarray> instance
        
    cdef c_from_param( self, object instance, object typeCode ):
        """simple function-based from_param"""
        cdef np.ndarray working = self.c_check_array( instance )
        cdef np.dtype targetType
        if typeCode:
            targetType = <np.dtype>(self.gl_constant_to_array[ typeCode ])
            if array_descr(instance) != targetType:
                raise CopyError(
                    """Array of type %r passed, required array of type %r""",
                    array_descr(instance).char, targetType.char,
                )                    
        if not PyArray_ISCARRAY_RO( instance ):
            raise CopyError(
                """from_param received a non-contiguous array! %s"""%(
                    working,
                )
            )
        return c_void_p(<size_t> PyArray_DATA(working))
    
    cdef c_dataPointer( self, object instance ):
        """Retrieve data-pointer directly"""
        return <size_t> PyArray_DATA(<np.ndarray>self.c_check_array( instance ))
    cdef c_zeros( self, object dims, object typeCode ):
        """Create an array initialized to zeros"""
        cdef np.ndarray c_dims
        try:
            c_dims = PyArray_ContiguousFromAny( 
                [int(x) for x in dims], np.NPY_INTP, 1,1 
            )
        except (ValueError,TypeError) as err:
            dims = (int(dims),)
            c_dims = PyArray_ContiguousFromAny( 
                dims, np.NPY_INTP, 1,1 
            )
        cdef np.dtype typecode = self.typeCodeToDtype( typeCode )
        Py_INCREF( typecode )
        cdef np.npy_intp ndims = PyArray_SIZE(c_dims)
        cdef np.ndarray result = PyArray_Zeros( <int>ndims, <np.npy_intp *>PyArray_DATA(c_dims), typecode, 0 )
        # Py_INCREF( result )
        return result
    cdef c_arraySize( self, object instance, object typeCode ):
        """Retrieve array size reference"""
        return (<np.ndarray>self.c_check_array( instance )).size
    cdef c_arrayByteCount( self, object instance ):
        """Given a data-value, calculate number of bytes required to represent"""
        return instance.nbytes
    cdef c_arrayToGLType( self, object instance ):
        """Given a value, guess OpenGL type of the corresponding pointer"""
        cdef np.ndarray value = self.c_check_array( instance )
        cdef object constant = self.array_to_gl_constant.get( array_descr(value) )
        if constant is None:
            raise TypeError(
                """Don't know GL type for array of type %r, known types: %s\nvalue:%s"""%(
                    array_descr(value), self.array_to_gl_constant.keys(), value,
                )
            )
        return constant
    cdef c_asArray( self, object instance, object typeCode ):
        """Retrieve the given value as a (contiguous) array of type typeCode"""
        cdef np.ndarray working
        cdef np.dtype typecode
        cdef int res
        if PyArray_CheckScalar(instance):
            Py_INCREF(instance)
            working = self.c_zeros((1,), typeCode)
            res = PyArray_FillWithScalar(working, instance)
            if res < -1:
                raise ValueError("Unable to fill new array with value %r (%s)"%(instance,instance.__class__))
        else:
            working = (<np.ndarray>self.c_check_array( instance ))
        if typeCode is None:
            typecode = array_descr(working)
        else:
            typecode = self.typeCodeToDtype( typeCode )
        return self.contiguous( working, typecode )
    cdef c_unitSize( self, object instance, typeCode ):
        """Retrieve last dimension of the array"""
        return PyArray_DIM(instance, PyArray_NDIM(instance)-1)
    cdef c_dimensions( self, object instance ):
        """Retrieve full set of dimensions for the array as tuple"""
        return instance.shape

    cdef np.dtype typeCodeToDtype( self, object typeCode ):
        """Convert type-code specification to a numpy dtype instance"""
        if isinstance( typeCode, np.dtype ):
            return typeCode
        elif isinstance( typeCode, (bytes,unicode) ):
            return np.dtype( typeCode )
        else:
            return self.gl_constant_to_array[ typeCode ]
    cdef np.ndarray contiguous( self, np.ndarray instance, np.dtype dtype ):
        """Ensure that this instance is a contiguous array
        
        :param instance: numpy instance to convert to contiguous array
        :param dtype: numpy dtype to which to convert the array
            IFF self.ERROR_ON_COPY then any required conversions 
            will raise a :class:`CopyError`
        
        :rtype: np.ndarray
        """
        if self.ERROR_ON_COPY:
            if not PyArray_ISCARRAY_RO( instance ):
                raise CopyError(
                    """Non-contiguous array passed""",
                    instance,
                )
            elif array_descr(instance) != dtype:
                raise CopyError(
                    """Array of type %r passed, required array of type %r""",
                    array_descr(instance).char, dtype.char,
                )
            # okay, so just return...
            return instance 
        else:
            # "convert" regardless (will return same instance if already contiguous)
            if PyArray_CheckScalar(instance):
                Py_INCREF( <object> instance )
                return PyArray_EnsureArray(instance)
            if not PyArray_ISCARRAY_RO( instance ) or array_descr(instance) != dtype:
                # TODO: make sure there's no way to segfault here 
                # Py_INCREF( <object> instance )
                Py_INCREF( <object> dtype )
                return PyArray_FromArray( 
                    instance, 
                    dtype, 
                    NPY_ARRAY_CARRAY|NPY_ARRAY_FORCECAST
                )

            else:
                return instance


# Cython numpy tutorial neglects to mention this AFAICS
# get segfaults without it
# import_array()



"""General pattern which does late-bound calls"""
#cython: language_level=3

cdef class LateBind:
    """Provides a __call__ which dispatches to self._finalCall

    When called without self._finalCall() makes a call to
    self.finalise() and then calls self._finalCall()
    """
    cdef object _finalCall
    def setFinalCall( self, object finalCall ):
        """Set our finalCall to the callable object given"""
        self._finalCall = finalCall

    def finalise( self ):
        """Finalise our target to our final callable object"""

    def __nonzero__(self):
        """Resolve our final call and check for empty/nonzero on it"""
        return bool(self.getFinalCall())
    def __bool__(self):
        """Resolve our final call and check for empty/nonzero on it"""
        return bool(self.getFinalCall())

    def getFinalCall( self ):
        """Retrieve and/or bind and retrieve final call"""
        if not self._finalCall:
            self._finalCall = self.finalise()
        return self._finalCall

    def __call__( self, *args, **named ):
        """Call self._finalCall, calling finalise() first if not already called

        There's actually *no* reason to unpack and repack the arguments,
        but unfortunately I don't know of a Cython syntax to specify
        that.
        """
        if not self._finalCall:
            self._finalCall = self.finalise()
        return self._finalCall( *args, **named )

cdef class Curry(object):
    """Provides a simple Curry which can bind (only) the first element"""
    cdef object wrapperFunction
    cdef object baseFunction
    def __init__( self, wrapperFunction, baseFunction ):
        """Stores self.wrapperFunction and self.baseFunction"""
        self.baseFunction = baseFunction
        self.wrapperFunction = wrapperFunction
    def __call__( self, *args, **named ):
        """returns self.wrapperFunction( self.baseFunction, *args, **named )"""
        return self.wrapperFunction( self.baseFunction, *args, **named )



"""Accelerator for None-as-an-array format handler operations"""
#cython: language_level=3
from OpenGL_accelerate.formathandler cimport FormatHandler

cdef class NoneHandler(FormatHandler):
	isOutput = False
	HANDLED_TYPES = (type(None),)
	
	cdef c_from_param( self, object instance, object typeCode ):
		"""simple function-based from_param"""
		return None
	cdef c_dataPointer( self, object instance ):
		"""Retrieve data-pointer directly"""
		return None
	cdef c_arraySize( self, object instance, object typeCode ):
		"""Retrieve array size reference"""
		return 0
	cdef c_arrayByteCount( self, object instance ):
		"""Given a data-value, calculate number of bytes required to represent"""
		return 0
	cdef c_asArray( self, object instance, object typeCode ):
		"""Retrieve the given value as a (contiguous) array of type typeCode"""
		return None
	cdef c_dimensions( self, object instance ):
		"""Retrieve full set of dimensions for the array as tuple"""
		return (0,)



import unittest, ctypes
from OpenGL.arrays import arraydatatype as adt
from OpenGL.arrays import vbo
from OpenGL import GL
from OpenGL._bytes import integer_types
import pytest
try:
    import numpy 
except ImportError:
    numpy = None
try:
    import OpenGL_accelerate
except ImportError:
    pytest.skip('Accelerate not installed, skipping', allow_module_level=True)

class _BaseTest( object ):
    array = None
    def setUp( self ):
        self.handler = adt.ArrayDatatype
        assert self.handler.isAccelerated
    def test_from_param( self ):
        p = self.handler.from_param( self.array )
        assert isinstance( p, ctypes.c_void_p )
    def test_dataPointer( self ):
        p = self.handler.dataPointer( self.array )
        assert isinstance( p, integer_types)
    def test_arraySize( self ):
        p = self.handler.arraySize( self.array )
        assert p == 6, p
    def test_arrayByteCount( self ):
        p = self.handler.arrayByteCount( self.array )
        assert p == 24, p
    def test_asArray( self ):
        p = self.handler.asArray( self.array )
        assert p is self.array 
    def test_unitSize( self ):
        p = self.handler.unitSize( self.array )
        assert p == 3, p
    def test_dimensions( self ):
        p = self.handler.dimensions( self.array )
        assert p == (2,3), p
    
    def test_arrayToGLType( self ):
        p = self.handler.arrayToGLType( self.array )
        assert p == GL.GL_FLOAT

# Skip if modifies the functions, which are *shared* between the 
# classes...
@pytest.mark.skipif( not numpy, reason="Numpy not available")
class TestNumpy( _BaseTest, unittest.TestCase ):
    def setUp( self ):
        super(TestNumpy,self).setUp()
        self.array = numpy.array( [[1,2,3],[4,5,6]],'f')
        handler = adt.ArrayDatatype.getHandler( self.array )
        handler.registerReturn( )

    def test_dataPointer( self ):
        p = self.handler.dataPointer( self.array )
        assert isinstance( p, integer_types)
        assert p == self.array.ctypes.data
    def test_zeros( self ):
        p = self.handler.zeros( (2,3,4), 'f' )
        assert p.shape == (2,3,4)
        assert p.dtype == numpy.float32
    def test_asArrayConvert( self ):
        p = self.handler.asArray( self.array, GL.GL_DOUBLE )
        assert p is not self.array 
        assert p.dtype == numpy.float64
        p = self.handler.asArray( self.array, 'd' )
        assert p is not self.array 
        assert p.dtype == numpy.float64
    def test_zeros_typed( self ):
        z = self.handler.zeros( (2,3,4), GL.GL_FLOAT)
        assert z.shape == (2,3,4)
        assert z.dtype == numpy.float32
    def test_downconvert( self ):
        p = self.handler.asArray( numpy.array( [1,2,3],'d'), GL.GL_FLOAT )
        assert p.dtype == numpy.float32
    def test_zeros_small( self ):
        z = self.handler.zeros( (0,), GL.GL_BYTE )
        assert z.dtype == numpy.byte, z

class TestVBO( _BaseTest, unittest.TestCase ):
    def setUp( self ):
        if numpy:
            self.array = vbo.VBO(numpy.array( [[1,2,3],[4,5,6]],'f'))
        else:
            self.array = vbo.VBO(adt.GLfloatArray.asArray([[1,2,3],[4,5,6]]))
        super(TestVBO,self).setUp()

class TestVBOOffset( _BaseTest, unittest.TestCase ):
    def setUp( self ):
        if numpy:
            self.array = vbo.VBO(numpy.array( [[1,2,3],[4,5,6]],'f')) + 12
        else:
            self.array = vbo.VBO(adt.GLfloatArray.asArray([[1,2,3],[4,5,6]])) + 12
        super(TestVBOOffset,self).setUp()
        
class TestNones( unittest.TestCase ):
    def setUp( self ):
        self.array = None
        self.handler = adt.ArrayDatatype
        assert self.handler.isAccelerated
    def test_from_param( self ):
        p = self.handler.from_param( self.array )
        assert p is None, p
    def test_dataPointer( self ):
        p = self.handler.dataPointer( self.array )
        assert p is None
    def test_asArray( self ):
        p = self.handler.asArray( self.array )
        assert p is self.array 
    def test_dimensions( self ):
        p = self.handler.dimensions( self.array )
        assert p == (0,), p



"""Cython memoryview implementation of buffer-api format handler"""
#cython: language_level=3

from ctypes import c_void_p
from OpenGL_accelerate.formathandler cimport FormatHandler
import traceback, weakref
from OpenGL.error import CopyError
from OpenGL._bytes import bytes,unicode

# Following is from the old python_buffer.pxd
cdef extern from "Python.h":

    cdef enum:
        PyBUF_SIMPLE,
        PyBUF_WRITABLE,
        PyBUF_WRITEABLE, # backwards compatability
        PyBUF_FORMAT,
        PyBUF_ND,
        PyBUF_STRIDES,
        PyBUF_C_CONTIGUOUS,
        PyBUF_F_CONTIGUOUS,
        PyBUF_ANY_CONTIGUOUS,
        PyBUF_INDIRECT,
        PyBUF_CONTIG,
        PyBUF_CONTIG_RO,
        PyBUF_STRIDED,
        PyBUF_STRIDED_RO,
        PyBUF_RECORDS,
        PyBUF_RECORDS_RO,
        PyBUF_FULL,
        PyBUF_FULL_RO,
        PyBUF_READ,
        PyBUF_WRITE,
        PyBUF_SHADOW

    bint PyObject_CheckBuffer(object obj)

    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags) except -1
    void PyBuffer_Release(object obj, object view)
    void* PyBuffer_GetPointer(Py_buffer *view, Py_ssize_t *indices)
    int PyObject_CopyToObject(object obj, void *buf, Py_ssize_t len, char fortran) except -1
    bint PyBuffer_IsContiguous(Py_buffer *view, char fort)
    
    object PyMemoryView_FromObject(object obj)
    object PyMemoryView_FromBuffer(Py_buffer *view)

    object PyMemoryView_GetContiguous(object obj, int buffertype, char order)
    
    bint PyMemoryView_Check(object obj)
    Py_buffer *PyMemoryView_GET_BUFFER(object obj)

cdef class MemoryviewHandler(FormatHandler):
    cdef public dict array_to_gl_constant
    cdef public dict gl_constant_to_array
    isOutput = True
    
    def __init__( self, ERROR_ON_COPY=None, a_to_gl=None ):
        if ERROR_ON_COPY is None:
            from OpenGL import _configflags
            ERROR_ON_COPY = _configflags.ERROR_ON_COPY
        if a_to_gl is None:
            try:
                from OpenGL.arrays._arrayconstants import ARRAY_TO_GL_TYPE_MAPPING
            except ImportError as err:
                raise RuntimeError( "Unable to load array constants" )
            a_to_gl = ARRAY_TO_GL_TYPE_MAPPING
        self.ERROR_ON_COPY = ERROR_ON_COPY
        self.array_to_gl_constant = a_to_gl

    cdef object c_as_memoryview( self, object instance ):
        """Ensure instance is a memory view instance or make it one"""
        if not PyMemoryView_Check( instance ):
            # TODO: respect no-copy flag!
            instance = PyMemoryView_GetContiguous( 
                instance, 
                PyBUF_STRIDES|PyBUF_FORMAT|PyBUF_C_CONTIGUOUS,
                'C'
            )
        return instance
    cdef c_from_param( self, object instance, object typeCode ):
        """simple function-based from_param"""
        if not PyMemoryView_Check( instance ):
            raise TypeError( '''Need a MemoryView instance in from_param (call asArray first)''' )
        return c_void_p( self.c_dataPointer( instance ) )
    cdef c_dataPointer( self, object instance ):
        """Retrieve data-pointer directly"""
        return <size_t> PyMemoryView_GET_BUFFER( self.c_as_memoryview( instance )).buf
    cdef c_zeros( self, object dims, object typeCode ):
        """Create an array initialized to zeros"""
        raise NotImplementedError( "Need to use a subclass to get zeros support" )
    cdef c_arraySize( self, object instance, object typeCode ):
        """Retrieve array size reference"""
        cdef Py_buffer * buffer
        buffer = PyMemoryView_GET_BUFFER( self.c_as_memoryview( instance ) )
        return buffer.len//buffer.itemsize
    cdef c_arrayByteCount( self, object instance ):
        """Given a data-value, calculate number of bytes required to represent"""
        return PyMemoryView_GET_BUFFER( self.c_as_memoryview( instance ) ).len
        
    cdef c_arrayToGLType( self, object instance ):
        """Given a value, guess OpenGL type of the corresponding pointer"""
        cdef Py_buffer * buffer
        buffer = PyMemoryView_GET_BUFFER( self.c_as_memoryview( instance ) )
        cdef object constant = self.array_to_gl_constant.get( buffer.format )
        if constant is None:
            if isinstance(buffer.format,bytes):
                constant = self.array_to_gl_constant.get(buffer.format.decode('utf-8'))
        if constant is None:
            raise TypeError(
                """Don't know GL type for array of type %r, known types: %s\nvalue:%s"""%(
                    buffer.format, self.array_to_gl_constant.keys(), buffer.format,
                )
            )
        return constant
        
    cdef c_asArray( self, object instance, object typeCode ):
        """Retrieve the given value as a (contiguous) array of type typeCode"""
        return self.c_as_memoryview( instance )
#        cdef Py_buffer * buffer
#        cdef object result
#        result = self.c_as_memoryview( instance )
#        buffer = PyMemoryView_GET_BUFFER( result )
#        actual_type = self.array_to_gl_constant.get( buffer.format, self.array_to_gl_constant )
#        if actual_type != typeCode:
#            raise TypeError( 'Incompatible type: %s, needed %s', buffer.format, typeCode )
#        return result
    cdef c_unitSize( self, object instance, typeCode ):
        """Retrieve last dimension of the array"""
        cdef Py_buffer * buffer
        view = self.c_as_memoryview( instance )
        buffer = PyMemoryView_GET_BUFFER( view )
        return buffer.shape[buffer.ndim-1]
    cdef c_dimensions( self, object instance ):
        """Retrieve full set of dimensions for the array as tuple"""
        cdef Py_buffer * buffer
        buffer = PyMemoryView_GET_BUFFER( self.c_as_memoryview( instance ) )
        cdef dim = 0
        cdef list result
        result = []
        for dim in range(buffer.ndim):
            result.append( buffer.shape[dim] )
        return result



"""Cython-coded VBO implementation"""
#cython: language_level=3
import ctypes, weakref
from OpenGL_accelerate.formathandler cimport FormatHandler
from OpenGL import error
from OpenGL._bytes import bytes,unicode

cdef extern from "Python.h":
    cdef void Py_XINCREF( object )
    cdef void Py_XDECREF( object )

_NULL = object()

cdef class VBO:
    """Instances can be passed into array-handling routines

    You can check for whether VBOs are supported by accessing the implementation
    attribute of the VBO, which will raise a RuntimeError if there is no available
    implementation.

    Attributes:

        int copied -- whether we are copied to back-end yet
        int created -- whether we've created a buffer yet
        unsigned int buffer -- our buffer once created
        object data -- our data-holding array-compatible object
        target -- our resolved GL constant target
        target_spec -- our (unresolved) GL constant specifier
        usage -- our resolved GL constant usage
        usage_spec -- our (unresolved) GL constant usage specifier
        _copy_segments -- slices of data to copy to back-end
        _I_ -- our implementation object
        arrayType -- our reference to arraydatatype.ArrayDatatype
    """
    cdef object __weakref__ # allow weak-referencing for cleanups...
    cdef public int copied
    cdef public int created
    cdef public unsigned int buffer
    cdef public object data
    cdef public int size
    cdef public object target
    cdef public object usage
    cdef public int resolved
    cdef public object target_spec # possible string definition
    cdef public object usage_spec # possible string definition
    cdef public list _copy_segments
    cdef public object _I_
    cdef public object arrayType
    _no_cache_ = True # do not cache in context data arrays
    def __init__(
        self, data, usage='GL_DYNAMIC_DRAW',
        target='GL_ARRAY_BUFFER',
        size=None,
    ):
        """Initialize the VBO

        data -- array-compatible data format object

        usage -- string/GLenum constant specifying streaming usage
            for the VBO, normal values:

                GL_STREAM_DRAW -- updated every frame from client to card
                GL_STREAM_COPY -- read and written from card each frame

                GL_STATIC_DRAW -- just written from client (once)
                GL_STATIC_READ -- just read from client
                GL_STATIC_COPY -- read and written from client (once)

                GL_DYNAMIC_DRAW -- updated from client every once in a while
                GL_DYNAMIC_COPY -- read and written from client every once in a while
                GL_DYNAMIC_READ -- read from the client every once in a while

        target -- string/GLenum constant specifying the role for the buffer
            normal values:

                GL_ARRAY_BUFFER -- vertex attribute values
                GL_ELEMENT_ARRAY_BUFFER -- vertex index values
                GL_PIXEL_PACK_BUFFER/GL_PIXEL_UNPACK_BUFFER -- client-side
                    image source/sink for image-manipulation operations.
        """
        from OpenGL.arrays.arraydatatype import ArrayDatatype
        self.arrayType = ArrayDatatype
        self.c_set_array( data, size )
        self.resolved = self.created = self.copied = False
        self.usage_spec = usage
        self.target_spec = target
        self._copy_segments = []
        self._I_ = None
    def __dealloc__( self ):
        """Deallocate our references"""
        if self.data is not None:
            Py_XDECREF( self.data )
            self.data = None
    @property
    def implementation( self ):
        """Retrieve our implementation reference"""
        return self.get_implementation()
    cdef get_implementation( self ):
        """C-level implementation retrieval"""
        if self._I_ is None:
            from OpenGL.arrays.vbo import get_implementation
            self._I_ = get_implementation()
        return self._I_
    cdef unsigned int c_resolve( self, value ):
        """Resolve string constant to constant"""
        if isinstance( value, (bytes,unicode)):
            return getattr(
                self.get_implementation(),
                self.get_implementation().basename( value )
            )
        return value
    def set_array( self, data, size=None ):
        """Update our entire array with new data"""
        return self.c_set_array( data, size )
    cdef c_set_array( self, data, size ):
        """Set our array pointer with incref/decref support"""
        if self.data is not None:
            Py_XDECREF( self.data )
        Py_XINCREF( data )
        self.data = data
        if size is None and self.data is not None:
            size = self.arrayType.arrayByteCount( self.data )
        self.size = size
        self.copied = False
    def __setitem__( self, slice, array):
        """Set slice of data on the array and vbo (if copied already)

        slice -- the Python slice object determining how the data should
            be copied into the vbo/array
        array -- something array-compatible that will be used as the
            source of the data, note that the data-format will have to
            be the same as the internal data-array to work properly, if
            not, the amount of data copied will be wrong.

        This is a reasonably complex operation, it has to have all sorts
        of state-aware changes to correctly map the source into the low-level
        OpenGL view of the buffer (which is just bytes as far as the GL
        is concerned).
        """
        if slice.step and not slice.step == 1:
            raise NotImplementedError( """Don't know how to map stepped arrays yet""" )
        # TODO: handle e.g. mapping character data into an integer data-set
        data = self.arrayType.asArray( array )
        start = (slice.start or 0)
        stop = (slice.stop or len(self.data))
        if start < 0:
            start += len(self.data)
            start = max((start,0))
        if stop < 0:
            stop += len(self.data)
            stop = max((stop,0))
        self.data[ slice ] = data
        if self.copied and self.created:
            if start-stop == len(self.data):
                # re-copy the whole data-set
                self.copied = False
            elif len(data):
                # now the fun part, we need to make the array match the
                # structure of the array we're going to copy into and make
                # the "size" parameter match the value we're going to copy in,
                # note that a 2D array (rather than a 1D array) may require
                # multiple mappings to copy into the memory area...

                # find the step size from the dimensions and base size...
                size = self.arrayType.arrayByteCount( self.data[0] )
                #baseSize = self.arrayType.unitSize( data )
                # now create the start and distance values...
                start *= size
                stop *= size
                # wait until the last moment (bind) to copy the data...
                self._copy_segments.append(
                    (start,(stop-start), data)
                )
    def __len__( self ):
        return len( self.data )
    def __getattr__( self, key ):
        if key not in ('data','usage','target','buffer', 'copied','_I_','implementation','_copy_segments' ):
            return getattr( self.data, key )
        else:
            raise AttributeError( key )
    def __int__( self ):
        """Get our VBO id"""
        if not self.buffer:
            self.create_buffers()
        return self.buffer
    def create_buffers( self ):
        """Create the internal buffer(s)"""
        assert not self.created, """Already created the buffer"""
        buffers = self.get_implementation().glGenBuffers(1)
        try:
            self.buffer = long( buffers )
        except (TypeError,ValueError) as err:
            self.buffer = buffers[0]
        self.target = self.c_resolve( self.target_spec )
        self.usage = self.c_resolve( self.usage_spec )
        self.created = True
        self.get_implementation()._DELETERS_[ id(self) ] = weakref.ref(
            # Cython instances can't have weakrefs, sigh...
            self,
            self.get_implementation().deleter( [self.buffer], id(self) )
        )
        return self.buffer
    def copy_data( self ):
        """Copy our data into the buffer on the GL side"""
        assert self.created, """Should do create_buffers before copy_data"""
        if self.copied:
            if self._copy_segments:
                while self._copy_segments:
                    start,size,data  = self._copy_segments.pop(0)
                    dataptr = self.arrayType.voidDataPointer( data )
                    self.get_implementation().glBufferSubData(self.target, start, size, dataptr)
        else:
            self.get_implementation().glBufferData(
                self.target,
                self.size,
                self.data,
                self.usage,
            )
            self.copied = True
    def delete( self ):
        """Delete this buffer explicitly"""
        if self.created:
            self.created = False
            try:
                self.get_implementation().glDeleteBuffers(1, ctypes.c_uint32( self.buffer))
            except (AttributeError, error.NullFunctionError) as err:
                pass
            try:
                self.get_implementation()._DELETERS_.pop( id(self))
            except KeyError as err:
                pass
    def bind( self ):
        """Bind this buffer for use in vertex calls"""
        if not self.created:
            buffer = self.create_buffers()
        self.get_implementation().glBindBuffer( self.target, self.buffer)
        self.copy_data()
    def unbind( self ):
        """Unbind the buffer (make normal array operations active)"""
        self.get_implementation().glBindBuffer( self.target,0 )
    def __add__( self, other ):
        """Add an integer to this VBO (offset)"""
        if hasattr( other, 'offset' ):
            other = other.offset
        assert isinstance( other, (int,long) ), """Only know how to add integer/long offsets"""
        return VBOOffset( self, other )
    cdef int check_live( self ):
        if self.data is _NULL:
            raise RuntimeError( """Attempting to use a deleted VBO""" )

    __enter__ = bind
    def __exit__( self, exc_type=None, exc_val=None, exc_tb=None ):
        """Context manager exit"""
        self.unbind()
        return False # do not supress exceptions...


cdef class VBOOffset:
    """Offset into a VBO instance

    This is just a convenience object that lets you say
    vbo + integer and have the value passed as the correct
    data-type instead of being interpreted as a pointer to
    an integer-array of the offset.
    """
    cdef public VBO vbo
    cdef public unsigned int offset
    def __cinit__( self, VBO vbo, unsigned int offset ):
        self.vbo = vbo
        Py_XINCREF( vbo )
        self.offset = offset
    def __dealloc__( self ):
        if self.vbo is not None:
            Py_XDECREF( self.vbo )
            self.vbo = None

    def __getattr__( self, key ):
        if key != 'vbo':
            return getattr( self.vbo, key )
        raise AttributeError( 'No %r key in VBOOffset'%(key,))
    def __add__( self, other ):
        if hasattr( other, 'offset' ):
            other = other.offset
        return VBOOffset( self.vbo, self.offset + other )
    cdef int check_live( self ):
        if self.vbo is not None:
            return self.vbo.check_live()
        else:
            raise RuntimeError( """Attempting to use offset into deleted VBO""" )

cdef class VBOHandler(FormatHandler):
    """Handles VBO instances passed in as array data"""
    cdef object vp0
    cdef object arrayType # import and use explicit reference...
    isOutput = False
    def __init__( self ):
        self.vp0 = ctypes.c_void_p( 0 )
        from OpenGL.arrays.arraydatatype import ArrayDatatype
        self.arrayType = ArrayDatatype
    cdef object c_dataPointer( self, object instance ):
        """Retrieve data-pointer directly"""
        (<VBO>instance).check_live()
        return 0
    cdef c_from_param( self, object instance, object typeCode ):
        """simple function-based from_param"""
        (<VBO>instance).check_live()
        return self.vp0
    cdef c_asArray( self, object instance, object typeCode ):
        """Retrieve the given value as a (contiguous) array of type typeCode"""
        (<VBO>instance).check_live()
        return instance
    cdef c_arrayByteCount( self, object instance ):
        """Given a data-value, calculate number of bytes required to represent"""
        (<VBO>instance).check_live()
        return self.arrayType.arrayByteCount( (<VBO>instance).data )
    cdef c_arrayToGLType( self, object instance ):
        """Given a value, guess OpenGL type of the corresponding pointer"""
        (<VBO>instance).check_live()
        return self.arrayType.arrayToGLType( (<VBO>instance).data )
    cdef c_arraySize( self, object instance, object typeCode ):
        """Retrieve array size reference"""
        (<VBO>instance).check_live()
        return self.arrayType.arraySize( (<VBO>instance).data )
    cdef c_unitSize( self, object instance, typeCode ):
        """Retrieve last dimension of the array"""
        (<VBO>instance).check_live()
        return self.arrayType.unitSize( (<VBO>instance).data )
    cdef c_dimensions( self, object instance ):
        """Retrieve full set of dimensions for the array as tuple"""
        (<VBO>instance).check_live()
        return self.arrayType.dimensions( (<VBO>instance).data )

cdef class VBOOffsetHandler(FormatHandler):
    cdef object arrayType # import and use explicit reference...
    isOutput = False
    def __init__( self ):
        from OpenGL.arrays.arraydatatype import ArrayDatatype
        self.arrayType = ArrayDatatype

    cdef object c_dataPointer( self, object instance ):
        """Retrieve data-pointer directly"""
        (<VBOOffset>instance).check_live()
        return (<VBOOffset>instance).offset
    cdef c_from_param( self, object instance, object typeCode ):
        """simple function-based from_param"""
        (<VBOOffset>instance).check_live()
        return ctypes.c_void_p( (<VBOOffset>instance).offset )
    cdef c_asArray( self, object instance, object typeCode ):
        """Retrieve the given value as a (contiguous) array of type typeCode"""
        (<VBOOffset>instance).check_live()
        return instance

    cdef c_arrayByteCount( self, object instance ):
        """Given a data-value, calculate number of bytes required to represent"""
        (<VBOOffset>instance).check_live()
        return self.arrayType.arrayByteCount( (<VBOOffset>instance).vbo.data )
    cdef c_arrayToGLType( self, object instance ):
        """Given a value, guess OpenGL type of the corresponding pointer"""
        (<VBOOffset>instance).check_live()
        return self.arrayType.arrayToGLType( (<VBOOffset>instance).vbo.data )
    cdef c_arraySize( self, object instance, object typeCode ):
        """Retrieve array size reference"""
        (<VBOOffset>instance).check_live()
        return self.arrayType.arraySize( (<VBOOffset>instance).vbo.data )
    cdef c_unitSize( self, object instance, typeCode ):
        """Retrieve last dimension of the array"""
        (<VBOOffset>instance).check_live()
        return self.arrayType.unitSize( (<VBOOffset>instance).vbo.data )
    cdef c_dimensions( self, object instance ):
        """Retrieve full set of dimensions for the array as tuple"""
        (<VBOOffset>instance).check_live()
        return self.arrayType.dimensions( (<VBOOffset>instance).vbo.data )



"""Cython-coded accelerators for the PyOpenGL wrapper

This package contains Cython accelerator modules which
attempt to speed up certain aspects of the PyOpenGL 3.x
wrapper mechanism.  The source code is part of the
PyOpenGL package and is built via the setupaccel.py
script in the top level of the PyOpenGL source package.
"""

__version__ = "3.1.9a2"
__version_tuple__ = (3, 1, 9)



"""Cython-coded implementation of wrapper module"""
#cython: language_level=3
import ctypes
from OpenGL import error
from OpenGL._null import NULL as _NULL

cdef extern from "Python.h":
    cdef object PyObject_Type( object )
    cdef object PyDict_GetItem( object, object )

cdef class cArgConverter:
    """C-level API definition for cConverter objects"""
    def __call__( self, tuple pyArgs, int index, object wrapper ):
        """Return pyArgs[self.index] or raise a ValueError"""
        return self.c_call( pyArgs, index, wrapper )
    cdef object c_call( self, tuple pyArgs, int index, object baseOperation ):
        """Implementation of the converter"""
cdef class pyArgConverter:
    """C-level API definition for pyArg converter objects"""
    def __call__( self, object incoming, object function, tuple arguments ):
        """Call our function on incoming"""
        return self.c_call( incoming, function, arguments )
    cdef object c_call( self, object incoming, object function, tuple arguments ):
        """Implementation would do something here"""
cdef class cArgumentConverter:
    """C-level API definition for cArgument convert objects"""
    def __call__( self, object element  ):
        """Call our function on incoming"""
        return self.c_call( element )
    cdef object c_call( self, object incoming ):
        """Implementation would do something here"""
cdef class returnConverter:
    """C-level API definition for cArgument convert objects"""
    def __call__( self, object result, object baseOperation, tuple pyArgs, tuple cArgs ):
        """Call our function on incoming"""
        return self.c_call( result, baseOperation, pyArgs, cArgs )
    cdef object c_call( self, object result, object baseOperation, tuple pyArgs, tuple cArgs ):
        """Implementation would do something here"""
    
    
cdef class CArgCalculatorElement:
    cdef object wrapper
    cdef long index 
    cdef int doCAPI
    cdef int callable
    cdef object converter 
    cdef cArgConverter c_converter
    def __init__( self, wrapper, index, converter ):
        self.wrapper = wrapper 
        self.index = index
        if isinstance( converter, cArgConverter ):
            self.c_converter = converter 
            self.doCAPI = True
            self.callable = True
        else:
            self.converter = converter
            self.doCAPI = False
            self.callable = hasattr( converter, '__call__' )
    def __call__( self, tuple pyArgs ):
        """If callable, call converter( pyArgs, index, wrapper ), else return converter"""
        return self.c_call( pyArgs )
    cdef object c_call( self, tuple pyArgs ):
        try:
            if self.doCAPI:
                return self.c_converter.c_call( pyArgs, self.index, self.wrapper )
            elif self.callable:
                return self.converter.__call__( pyArgs, self.index, self.wrapper )
        except Exception as err:
            err.args += ( self.index, self.wrapper )
            raise
        return self.converter

cdef class CArgCalculator:
    """C-coded version of the c-arg calculator pattern"""
    cdef list mapping
    def __init__( 
        self,
        wrapper,
        cConverters
    ):
        self.mapping = [
            CArgCalculatorElement(wrapper,i,converter)
            for (i,converter) in enumerate( cConverters )
        ]
    def __call__( self, tuple pyArgs ):
        return self.c_call( pyArgs )
    cdef list c_call( self, tuple pyArgs ):
        return [
            (<CArgCalculatorElement> calc).c_call( pyArgs )
            for calc in self.mapping
        ]
        

cdef class PyArgCalculatorElement:
    cdef object wrapper
    cdef long index 
    cdef int doCAPI
    cdef object converter 
    cdef pyArgConverter c_converter
    def __init__( self, wrapper, index, converter ):
        self.wrapper = wrapper 
        self.index = index 
        if isinstance( converter, pyArgConverter ):
            self.doCAPI = True 
            self.c_converter = converter
        else:
            self.doCAPI = False 
            self.converter = converter 
    def __call__( self, tuple args ):
        """If converter is not None, call converter( pyArgs[self.index],  wrapper, pyArgs ), else return args[self.index]"""
        return self.c_call( args )
    cdef object c_call( self, tuple args ):
        try:
            if self.doCAPI:
                return self.c_converter.c_call(
                    args[self.index], self.wrapper, args,
                )
            else:
                if self.converter is None:
                    return args[self.index]
                if len(args) <= self.index:
                    return _NULL
                return self.converter( 
                    args[self.index], self.wrapper, args 
                )
        except Exception as err:
            if hasattr( err, 'args' ):
                err.args += ( self.converter, )
            raise

cdef class PyArgCalculator:
    """C-coded version of the py-arg calculator pattern"""
    cdef list mapping
    cdef int length
    cdef object wrapper
    def __init__( 
        self,
        wrapper,
        pyConverters
    ):
        self.wrapper = wrapper
        self.mapping = [
            PyArgCalculatorElement(self,i,converter)
            for (i,converter) in enumerate( pyConverters )
        ]
        self.length = len([ p for p in pyConverters if not getattr(p,'optional',False)])
        
    def __call__( self, tuple args ):
        return self.c_call( args )
    cdef list c_call( self, tuple args ):
        if self.length > len(args):
            raise ValueError(
                """%s requires %r arguments (%s), received %s: %r"""%(
                    self.wrapper.wrappedOperation.__name__,
                    self.length,
                    ", ".join( self.wrapper.pyConverterNames ),
                    len(args),
                    args
                )
            )
        return [
            (<PyArgCalculatorElement> calc).c_call( args )
            for calc in self.mapping
        ]

cdef class CArgumentCalculator:
    cdef list cResolvers
    cdef int resolver_length
    def __init__( self, cResolvers ):
        self.cResolvers = cResolvers
        self.resolver_length = len(self.cResolvers)
    def __call__( self, tuple cArgs ):
        return self.c_call( cArgs )
    cdef list c_call( self, tuple cArgs ):
        cdef int i
        cdef int resolver_length
        cdef object converter
        if len(cArgs) != self.resolver_length:
            raise TypeError(
                """Expected %s C arguments for resolution, got %s"""%(
                    self.resolver_length,len(cArgs)
                )
            )
        result = [None]*self.resolver_length
        for i in range( self.resolver_length ):
            converter = self.cResolvers[i]
            if converter is None:
                result[i] = cArgs[i]
            elif isinstance( converter, cArgumentConverter ):
                # TODO: should pre-calculate this, isinstance is a very 
                # heavy operation.
                try:
                    result[i] = (<cArgumentConverter>converter).c_call( cArgs[i] )
                except Exception as err:
                    err.args += (converter,)
                    raise
            else:
                try:
                    result[i] = converter( cArgs[i] )
                except Exception as err:
                    err.args += (converter,)
                    raise
        return result

cdef class CallFuncPyConverter( pyArgConverter ):
    """PyConverter that takes a callable and calls it on incoming"""
    cdef object function
    def __init__( self, function ):
        """Store the function"""
        self.function = function 
    cdef object c_call( self, object incoming, object function, tuple arguments ):
        """Call our function on incoming"""
        return self.function( incoming )
    def __repr__( self ):
        """Have a useful representation"""
        return '%s( %r )'%(
            self.__class__.__name__,
            self.function,
        )
cdef class DefaultCConverter(cArgConverter):
    cdef int index
    def __init__( self, index ):
        """Just store index for future access"""
        self.index = index 
    cdef object c_call( self, tuple pyArgs, int index, object baseOperation ):
        try:
            return pyArgs[ self.index ]
        except IndexError as err:
            raise ValueError(
                """Expected parameter index %r, but pyArgs only length %s"""%(
                self.index,
                len(pyArgs )
            ))
    def __repr__( self ):
        return """%s( %r )"""%(
            self.__class__.__name__,
            self.index,
        )

cdef class Wrapper:
    """C-coded most-generic form of the wrapper's core function
    
    Applies a series of wrapper stages to the core function in order 
    to expand and/or convert parameters into formats which are compatible
    with the underlying (ctypes) library.  Requires the use of the 
    Calculator objects defined later in this module in order to provide 
    efficient operations.
    """
    cdef CArgCalculator calculate_cArgs
    cdef CArgumentCalculator calculate_cArguments
    cdef PyArgCalculator calculate_pyArgs
    
    cdef public object wrappedOperation, storeValues,returnValues
    cdef int doPyargs, doCargs, doCarguments, doStoreValues, doReturnValues
    cdef returnConverter c_returnValues
    cdef int doCReturnAPI
    def __init__( 
        self, wrappedOperation,
        calculate_pyArgs=None, calculate_cArgs=None,
        calculate_cArguments=None, 
        storeValues=None, returnValues=None,
    ):
        if calculate_pyArgs is not None:
            if not isinstance( calculate_pyArgs, PyArgCalculator ):
                calculate_pyArgs = PyArgCalculator( self, calculate_pyArgs )
            self.calculate_pyArgs = calculate_pyArgs
            self.doPyargs = True
        else:
            self.doPyargs = False
        if calculate_cArgs is not None:
            if not isinstance( calculate_cArgs, CArgCalculator ):
                calculate_cArgs = CArgCalculator( self, calculate_cArgs )
            self.calculate_cArgs = calculate_cArgs
            self.doCargs = True 
        else:
            self.doCargs = False
        if calculate_cArguments is not None:
            if not isinstance( calculate_cArguments, CArgumentCalculator ):
                calculate_cArguments = CArgumentCalculator( self, calculate_cArguments )
            self.calculate_cArguments = calculate_cArguments
            self.doCarguments = True
        else:
            self.doCarguments = False
        self.wrappedOperation = wrappedOperation
        self.storeValues = storeValues
        if isinstance( returnValues, returnConverter ):
            self.doCReturnAPI = True 
            self.doReturnValues = True
            self.c_returnValues = returnValues
        elif returnValues is None:
            self.doReturnValues = False
        else:
            self.doCReturnAPI = False 
            self.doReturnValues = True 
            self.returnValues = returnValues
    
    def __call__( self, *args ):
        cdef tuple pyArgs, cArgs, cArguments
        if self.doPyargs:
            pyArgs = tuple(self.calculate_pyArgs.c_call( args ))
        else:
            pyArgs = args 
        if self.doCargs:
            cArgs = tuple(self.calculate_cArgs.c_call( pyArgs ))
        else:
            cArgs = pyArgs
        if self.doCarguments:
            cArguments = tuple(self.calculate_cArguments( cArgs ))
        else:
            cArguments = cArgs
        try:
            result = self.wrappedOperation( *cArguments )
        except (ctypes.ArgumentError,TypeError,AttributeError) as err:
            err.args = err.args + (cArguments,)
            raise err
        except error.GLError as err:
            err.cArgs = cArgs 
            err.pyArgs = pyArgs
            raise err
        # handle storage of persistent argument values...
        if self.storeValues is not None:
            self.storeValues(
                result,
                self,
                pyArgs,
                cArgs,
            )
        if self.doReturnValues:
            if self.doCReturnAPI:
                return self.c_returnValues.c_call(
                    result, self, pyArgs, cArgs,
                )
            else:
                try:
                    return self.returnValues(
                        result,
                        self,
                        pyArgs,
                        cArgs,
                    )
                except Exception as err:
                    err.args += (self.returnValues,)
                    raise
        else:
            return result


cdef class getPyArgsName(cArgConverter):
    """CConverter returning named Python argument
    
    Intended for use in cConverters, the function returned 
    retrieves the named pyArg and returns it when called.
    """
    cdef public unsigned int index
    cdef public str name
    def __init__( self, str name ):
        self.name = name
    def finalise( self, wrapper ):
        self.index = wrapper.pyArgIndex( self.name )
    cdef object c_call( self, tuple pyArgs, int index, object baseOperation ):
        return pyArgs[ self.index ]
    def __repr__( self ):
        return """%s( %r )"""%(
            self.__class__.__name__,
            self.name,
        )

cdef class returnPyArgument(returnConverter):
    """ReturnValues returning the named pyArgs value"""
    cdef public unsigned int index
    cdef public str name
    def __init__( self, str name ):
        self.name = name 
    def finalise( self, wrapper ):
        self.index = wrapper.pyArgIndex( self.name )
    cdef c_call( self, object result, object baseOperation, tuple pyArgs, tuple cArgs ):
        """Retrieve pyArgs[ self.index ]"""
        return pyArgs[self.index]
    def __repr__( self ):
        return """%s( %r )"""%(
            self.__class__.__name__,
            self.name,
        )
cdef class returnPyArgumentIndex(returnConverter):
    cdef public unsigned int index
    def __init__( self, int index ):
        self.index = index
    def finalise( self, wrapper ):
        """No finalisation required"""
    cdef c_call( self, object result, object baseOperation, tuple pyArgs, tuple cArgs ):
        """Retrieve pyArgs[ self.index ]"""
        return pyArgs[self.index]
    def __repr__( self ):
        return """%s( %r )"""%(
            self.__class__.__name__,
            self.index,
        )
cdef class returnCArgument(returnConverter):
    """ReturnValues returning the named pyArgs value"""
    cdef public unsigned int index
    cdef public str name
    def __init__( self, str name ):
        self.name = name 
    def finalise( self, wrapper ):
        self.index = wrapper.cArgIndex( self.name )
    cdef c_call( self, object result, object baseOperation, tuple pyArgs, tuple cArgs ):
        """Retrieve cArgs[ self.index ]"""
        return cArgs[self.index]
    def __repr__( self ):
        return """%s( %r )"""%(
            self.__class__.__name__,
            self.name,
        )
    
cdef class MultiReturn(object):
    cdef public list children
    def __init__(self,*children):
        self.children = list(children)
    def append(self, child ):
        self.children.append( child )
    def __call__(self,*args,**named):
        cdef list result = []
        for child in self.children:
            try:
                result.append( child(*args,**named) )
            except Exception as err:
                err.args += ( child, args, named )
                raise
        return result
    



"""Importable Cython declarations for wrapper module"""
#cython: language_level=3

cdef class cArgConverter:
	cdef object c_call( self, tuple pyArgs, int index, object baseOperation )
cdef class pyArgConverter:
	cdef object c_call( self, object incoming, object function, tuple arguments )
cdef class cArgumentConverter:
	cdef object c_call( self, object incoming )
cdef class returnConverter:
	cdef object c_call( self, object result, object baseOperation, tuple pyArgs, tuple cArgs )



import unittest, ctypes
try:
    import numpy 
    from OpenGL_accelerate import numpy_formathandler as npf
except ImportError:
    numpy = None
    npf = None
try:
    from OpenGL_accelerate import buffers_formathandler as bpf
except ImportError:
    bpf = None
from OpenGL import error
from OpenGL import GL
from OpenGL._bytes import integer_types
from OpenGL._configflags import ERROR_ON_COPY
import pytest
pytestmark = pytest.mark.skipif(not numpy, reason="No numpy installed in order to run tests")

class _AccelArray( object ):
    handler_class = None
    def setUp( self ):
        self.array = numpy.array( [[1,2,3],[4,5,6]],'f')
        self.handler = self.handler_class()
    def test_from_param( self ):
        p = self.handler.from_param( self.handler.asArray(self.array ))
        assert isinstance( p, ctypes.c_void_p )
    def test_dataPointer( self ):
        p = self.handler.dataPointer( self.array )
        assert isinstance( p, integer_types)
        assert p == self.array.ctypes.data
    def test_arraySize( self ):
        p = self.handler.arraySize( self.array )
        assert p == 6, p
    def test_arrayByteCount( self ):
        p = self.handler.arrayByteCount( self.array )
        assert p == 24, p
    def test_unitSize( self ):
        p = self.handler.unitSize( self.array )
        assert p == 3, p
    def test_dimensions( self ):
        p = self.handler.dimensions( self.array )
        assert p == (2,3), p
    
    def test_arrayToGLType( self ):
        p = self.handler.arrayToGLType( self.array )
        assert p == GL.GL_FLOAT
        
    

@pytest.mark.skipif(not npf,reason="No numpy native format handler available")
class TestNumpyNative(_AccelArray,unittest.TestCase):
    handler_class = getattr(npf,'NumpyHandler',None)
    def setUp(self):
        super(TestNumpyNative,self).setUp()
        self.eoc_handler = self.handler_class( True )
        
    def test_asArray( self ):
        p = self.handler.asArray( self.array )
        assert p is self.array 
    def test_downconvert( self ):
        p = self.handler.asArray( numpy.array( [1,2,3],'d'), GL.GL_FLOAT )
        assert p.dtype == numpy.float32
    def test_zeros_constant( self ):
        z = self.handler.zeros( (2,3,4), GL.GL_FLOAT)
        assert z.shape == (2,3,4)
        assert z.dtype == numpy.float32
    def test_zeros( self ):
        p = self.handler.zeros( (2,3,4), 'f' )
        assert p.shape == (2,3,4)
        assert p.dtype == numpy.float32
    def test_asArrayCopy( self ):
        a2 = self.array[:,0]
        assert not a2.flags.contiguous 
        self.assertRaises(
            error.CopyError,
            self.eoc_handler.asArray,
                a2
        )
    def test_asArrayConvert( self ):
        self.failUnlessRaises(
            error.CopyError,
            self.eoc_handler.asArray,
                self.array, GL.GL_DOUBLE 
        )
    def test_asArrayConvert( self ):
        p = self.handler.asArray( self.array, GL.GL_DOUBLE )
        assert p is not self.array 
        assert p.dtype == numpy.float64
        p = self.handler.asArray( self.array, 'd' )
        assert p is not self.array 
        assert p.dtype == numpy.float64


@pytest.mark.skipif(not npf,reason="No numpy native format handler available")
class TestBufferAPI(_AccelArray,unittest.TestCase):
    handler_class = getattr(bpf,'MemoryviewHandler',None)



"""Cython-coded GL-error-check module"""
#cython: language_level=3
from OpenGL import _configflags

cdef class _ErrorChecker:
    """Global error-checking object
    
    This error checker also includes "safeGetError" functionality,
    that is, it allows for checking for context validity as well 
    as for glBegin/glEnd checking.
    """
    cdef public int doChecks
    cdef public int checkContext
    cdef public object _isValid
    cdef public object _getErrors
    cdef public object _errorClass
    cdef public int _noErrorResult 
    
    def __init__( self, platform, baseOperation, noErrorResult=0, errorClass=None ):
        """Initialize from a platform module/reference"""
        self._isValid = platform.CurrentContextIsValid
        self._getErrors = baseOperation
        self._noErrorResult = noErrorResult
        self._errorClass = errorClass
        
        self.doChecks = bool( _configflags.ERROR_CHECKING and self._getErrors )
        self.checkContext = _configflags.CONTEXT_CHECKING
    
    def glCheckError( 
        self,
        result,
        baseOperation=None,
        cArguments=None,
    ):
        """Base GL Error checker compatible with new ctypes errcheck protocol
        
        This function will raise a GLError with just the calling information
        available at the C-calling level, i.e. the error code, cArguments,
        baseOperation and result.  Higher-level code is responsible for any 
        extra annotations.
        
        Note:
            glCheckError relies on glBegin/glEnd interactions to 
            prevent glGetError being called during a glBegin/glEnd 
            sequence.  If you are calling glBegin/glEnd in C you 
            should call onBegin and onEnd appropriately.
        """
        cdef int err
        if self.doChecks:
            if self.checkContext:
                if not self._isValid():
                    return 
            err = self._getErrors()
            if err != self._noErrorResult:
                if self._errorClass is None:
                    # circular import here otherwise
                    from OpenGL.error import GLError
                    self._errorClass = GLError
                raise self._errorClass(
                    err,
                    result,
                    cArguments = cArguments,
                    baseOperation = baseOperation,
                )
        return result
    def onBegin( self, target=None ):
        """Called by glBegin to record the fact that glGetError won't work"""
        self.doChecks = False
    def onEnd( self, target=None ):
        """Called by glEnd to record the fact that glGetError will work"""
        self.doChecks = bool( _configflags.ERROR_CHECKING and self._getErrors )
    def check( self ):
        return self.glCheckError( None, self._getErrors, [] )



"""Cython import description for formathandler types"""
#cython: language_level=3

cdef class FormatHandler:
	cdef public int ERROR_ON_COPY
	cdef object c_from_param( self, object instance, object typeCode)
	cdef object c_dataPointer( self, object instance )
	cdef c_zeros( self, object dims, object typeCode )
	cdef c_arraySize( self, object instance, object typeCode )
	cdef c_arrayByteCount( self, object instance )
	cdef c_arrayToGLType( self, object value )
	cdef c_asArray( self, object instance, object typeCode)
	cdef c_unitSize( self, object instance, typeCode )
	cdef c_dimensions( self, object instance)



#!/usr/bin/env python
"""Builds accelleration functions for PyOpenGL
"""
import sys
from setuptools import setup, Extension

try:
    from Cython.Distutils import build_ext
except ImportError:
    have_cython = False
    build_ext = False
else:
    have_cython = True

import sys, os

HERE = os.path.normpath(os.path.abspath(os.path.dirname(__file__)))

extensions = []


def cython_extension(
    name,
    include_dirs=(),
):
    """Create a cython extension object"""
    filenames = "%(name)s.c" % locals(), "%(name)s.pyx" % locals()
    filename = filenames[bool(have_cython)]
    return Extension(
        "OpenGL_accelerate.%(name)s" % locals(),
        [
            os.path.join("src", filename),
        ],
        include_dirs=[
            os.path.join(HERE, "src"),
            HERE,
        ]
        + list(include_dirs),
        define_macros=[
            ('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION'),
        ],
        compiler_directives={'language_level': "3"} if have_cython else {},
    )


extensions.extend(
    [
        cython_extension("wrapper"),
        cython_extension("formathandler"),
        cython_extension("arraydatatype"),
        cython_extension("errorchecker"),
        cython_extension("vbo"),
        cython_extension("nones_formathandler"),
        cython_extension("latebind"),
    ]
)
if sys.version_info[:2] >= (2, 7):
    extensions.extend(
        [
            cython_extension("buffers_formathandler"),
        ]
    )

try:
    import numpy
except ImportError:
    sys.stderr.write("""Unable to import numpy, skipping numpy extension building\n""")
else:
    if hasattr(numpy, "get_include"):
        includeDirectories = [
            numpy.get_include(),
        ]
    else:
        includeDirectories = [
            os.path.join(
                os.path.dirname(numpy.__file__),
                "core",
                "include",
            ),
        ]
    extensions.append(
        cython_extension(
            "numpy_formathandler",
            includeDirectories,
        )
    )

if (  # Prevents running of setup during code introspection imports
    __name__ == "__main__"
):
    # Workaround for Broken apple Python build params echoed in distutils
    # Approach taken from the PyMongo driver. OS-X Python builds were created with
    # non-existent flag, and distutils passes those flags to the extension
    # build, newer clangs now treat those flags as errors rather than warnings.
    import platform, sys

    if sys.platform == "darwin" and "clang" in platform.python_compiler().lower():
        from distutils.sysconfig import get_config_vars

        res = get_config_vars()
        for key in ("CFLAGS", "PY_CFLAGS"):
            if key in res:
                res[key] = res[key].replace("-mno-fused-madd", "")

    extraArguments = {}
    ### Now the actual set up call
    if have_cython:
        extraArguments["cmdclass"] = {
            "build_ext": build_ext,
        }
    setup(
        build_requires=['cython'],
        options={
            "sdist": {
                "formats": ["gztar"],
                "force_manifest": True,
            },
        },
        ext_modules=extensions,
        **extraArguments
    )



"""Cython-coded base FormatHandler implementation"""
#cython: language_level=3

cdef class FormatHandler:
	"""PyOpenGL "Format" plugin implementation
	
	PyOpenGL 3.x allows for multiple Formats for data arrays,
	a FormatHandler provides generic API entry points which 
	allow PyOpenGL to manipulate/interact with data arrays
	in the given format.
	
	This is the C-level accelerator module which allows
	for much faster access to the accelerator by the C-level
	wrapper classes.  This FormatHandler type should only 
	be used as a base class by those FormatHandlers which 
	provide C-level API entry points, you should *not*
	sub-class it in Python.
	"""	
	isOutput = False
	isAccelerated = True
	HANDLED_TYPES = ()
	
	def __init__( self):
		"""Initialize the format handler"""
	def register( self, types=None ):
		"""Register this class as handler for given set of types"""
		from OpenGL.arrays.arraydatatype import ArrayDatatype
		ArrayDatatype.getRegistry().register( self, types )
	def registerReturn( self ):
		"""Register this handler as the default return-type handler"""
		from OpenGL.arrays.arraydatatype import ArrayDatatype
		ArrayDatatype.getRegistry().registerReturn( self )
	def registerEquivalent( self, typ, base ):
		"""Register a sub-class for handling as the base-type"""
		
	def from_param( self, object instance, object typeCode = None ):
		"""simple function-based from_param"""
		return self.c_from_param( instance, typeCode )
	cdef object c_from_param( self, object instance, object typeCode ):
		"""C-API for from_param"""
		raise NotImplementedError( """%s does not define from_param method"""%(self,))
	def dataPointer( self, object instance ):
		"""Retrieve data-pointer directly"""
		return self.c_dataPointer( instance )
	cdef object c_dataPointer( self, object instance ):
		"""Retrieve data-pointer directly"""
		raise NotImplementedError( """%s does not define dataPointer method"""%(self,))
	def zeros( self, object dims, object typeCode ):
		"""Create an array initialized to zeros"""
		return self.c_zeros( dims, typeCode )
	cdef c_zeros( self, object dims, object typeCode ):
		"""Create an array initialized to zeros"""
		raise NotImplementedError( """%s does not define zeros method"""%(self,))
	def arraySize( self, object instance, object typeCode=None ):
		"""Retrieve array size reference"""
		return self.c_arraySize( instance, typeCode )
	cdef c_arraySize( self, object instance, object typeCode ):
		"""Retrieve array size reference"""
		raise NotImplementedError( """%s does not define arraySize method"""%(self,))
	def arrayByteCount( self, object instance ):
		"""Given a data-value, calculate number of bytes required to represent"""
		return self.c_arrayByteCount( instance )
	cdef c_arrayByteCount( self, object instance ):
		"""Given a data-value, calculate number of bytes required to represent"""
		raise NotImplementedError( """%s does not define arrayByteCount method"""%(self,))
	def arrayToGLType( self, object value ):
		"""Given a value, guess OpenGL type of the corresponding pointer"""
		return self.c_arrayToGLType( value )
	cdef c_arrayToGLType( self, object value ):
		"""Given a value, guess OpenGL type of the corresponding pointer"""
		raise NotImplementedError( """%s does not define arrayToGLType method"""%(self,))
	def asArray( self, object instance, object typeCode = None ):
		"""Retrieve the given value as a (contiguous) array of type typeCode"""
		return self.c_asArray( instance, typeCode )
	cdef c_asArray( self, object instance, object typeCode ):
		"""Retrieve the given value as a (contiguous) array of type typeCode"""
		raise NotImplementedError( """%s does not define asArray method"""%(self,))
	def unitSize( self, object instance, typeCode=None ):
		"""Retrieve last dimension of the array"""
		return self.c_unitSize( instance, typeCode )
	cdef c_unitSize( self, object instance, typeCode ):
		"""Retrieve last dimension of the array"""
		raise NotImplementedError( """%s does not define unitSize method"""%(self,))
	def dimensions( self, object instance ):
		"""Retrieve full set of dimensions for the array as tuple"""
		return tuple(self.c_dimensions( instance ))
	cdef c_dimensions( self, object instance ):
		"""Retrieve full set of dimensions for the array as tuple"""
		raise NotImplementedError( """%s does not define dimensions method"""%(self,))



"""Cython-coded Array-handling accelerator module"""
#cython: language_level=3

import ctypes
import OpenGL
from OpenGL._null import NULL as _NULL
from OpenGL import plugins
from OpenGL_accelerate.wrapper cimport cArgConverter, pyArgConverter, returnConverter
from OpenGL_accelerate.formathandler cimport FormatHandler

cdef extern from "Python.h":
    cdef object PyObject_Type( object )
    cdef object PyDict_GetItem( object, object )

cdef class HandlerRegistry:
    """C-coded registry of format handlers for array data-formats"""
    cdef dict registry
    cdef object match
    cdef object output_handler
    cdef object preferredOutput
    GENERIC_OUTPUT_PREFERENCES = ['numpy','ctypesarrays']
    cdef object all_output_handlers
    def __init__( self, plugin_match ):
        self.registry = {}
        self.match = plugin_match
        self.output_handler = None 
        self.preferredOutput = None
        self.all_output_handlers = []
    def __setitem__( self,key,value ):
        self.registry[key] = value
    def __call__( self, value ):
        return self.c_lookup( value )
    
    cdef object c_lookup( self, object value ):
        """C-level lookup of handler for given value"""
        cdef object typ, handler,base
        try:
            typ = value.__class__
        except AttributeError as err:
            typ = PyObject_Type(value)
        handler = self.registry.get( typ )
        if not handler:
            if hasattr( typ, '__mro__' ):
                for base in typ.__mro__:
                    handler = self.registry.get( base )
                    if not handler:
                        plugin = self.match( base )
                        if plugin:
                            cls = plugin.load()
                            if cls:
                                handler=cls()
                    if handler:
                        self.registry[ typ ] = handler
                        if hasattr( handler, 'registerEquivalent' ):
                            handler.registerEquivalent( typ, base )
                        return handler
            raise TypeError(
                """No array-type handler for type %r (value: %s) registered"""%(
                    typ, repr(value)[:50]
                )
            )
        return handler
    
    cdef object c_get_output_handler( self ):
        """Fast-path lookup for output handler object"""
        if self.output_handler is None:
            if self.preferredOutput is not None:
                self.output_handler = self.c_handler_by_plugin_name( self.preferredOutput )
            if not self.output_handler:
                for preferred in self.GENERIC_OUTPUT_PREFERENCES:
                    self.output_handler = self.c_handler_by_plugin_name( preferred )
                    if self.output_handler:
                        break
            if not self.output_handler:
                raise RuntimeError(
                    """Unable to find any output handler at all (not even ctypes/numpy ones!)"""
                )
        return self.output_handler
    cdef object c_handler_by_plugin_name( self, str name ):
        plugin = plugins.FormatHandler.by_name( name )
        try:
            handler = plugin.load()
            if handler:
                return handler()
        except ImportError as err:
            return None
    
    def register( self, handler, types=None ):
        """Register this class as handler for given set of types"""
        if not isinstance( types, (list,tuple)):
            types = [ types ]
        for type in types:
            self.registry[ type ] = handler
        if handler.isOutput:
            self.all_output_handlers.append( handler )
        
    def registerReturn( self, handler ):
        """Register this handler as the default return-type handler"""
        if isinstance( handler, (str,unicode)):
            self.preferredOutput = handler 
            self.output_handler = None
        else:
            self.preferredOutput = None 
            self.output_handler = handler

cdef HandlerRegistry GLOBAL_REGISTRY = HandlerRegistry(plugins.FormatHandler.match)

cdef class ArrayDatatype:
    """Mix-in for array datatype classes
    
    The ArrayDatatype marker essentially is used to mark a particular argument
    as having an "array" type, which means that it is eligible for handling 
    via the arrays sub-package and its registered handlers.
    """
    cdef HandlerRegistry handler 
    cdef public object typeConstant
    cdef public object baseType
    isAccelerated = True
    def __init__( self, typeConstant=None, baseType=None ):
        """Initialize, grabbing our handler registry"""
        self.typeConstant = typeConstant
        self.baseType = baseType
        self.handler = GLOBAL_REGISTRY
    
    def getRegistry( self ):
        """Get our handler registry"""
        return self.handler 
    def getHandler( self, value ):
        """Retrieve FormatHandler for given value 
        
        This method is replaced by the FormatHandler registry
        once the registry is initialized...
        """
        return self.handler.c_lookup( value )
    
    def from_param( self, object value, object typeConstant=None ):
        """Given a value in a known data-pointer type, convert to a ctypes pointer"""
        handler = self.handler.c_lookup( value )
        if isinstance( handler, FormatHandler ):
            return (<FormatHandler>handler).c_from_param( 
                value, typeConstant or self.typeConstant 
            )
        return handler.from_param( value, typeConstant or self.typeConstant )
    
    def __call__( self, object value ):
        """We cannot simply reference from_param as under Python 2.7 that makes us non-callable"""
        return self.from_param( value )
    
    def dataPointer( self, value ):
        """Given a value in a known data-pointer type, return long for pointer"""
        handler = self.handler.c_lookup( value )
        if isinstance( handler, FormatHandler ):
            return (<FormatHandler>handler).c_dataPointer( 
                value
            )
        return handler.dataPointer( value )
    def voidDataPointer( self, value ):
        """Given value in a known data-pointer type, return void_p for pointer"""
        pointer = self.dataPointer( value )
        try:
            return ctypes.c_void_p(pointer)
        except TypeError as err:
            return pointer
    def typedPointer( self, value ):
        """Return a pointer-to-base-type pointer for given value"""
        return ctypes.cast( 
            self.dataPointer(value), 
            ctypes.POINTER( self.baseType )
        )
    def asArray( self, value, typeCode=None ):
        """Given a value, convert to preferred array representation"""
        if typeCode is None:
            typeCode = self.typeConstant
        handler = self.handler.c_lookup( value )
        if isinstance( handler, FormatHandler ):
            return (<FormatHandler>handler).c_asArray( 
                value, typeCode 
            )
        return handler.asArray( 
            value, typeCode
        )
    def arrayToGLType( self, value ):
        """Given a data-value, guess the OpenGL type of the corresponding pointer
        
        Note: this is not currently used in PyOpenGL and may be removed 
        eventually.
        """
        return self.handler.c_lookup( value ).arrayToGLType( value )
    def arraySize( self, value, typeCode = None ):
        """Given a data-value, calculate dimensions for the array (number-of-units)"""
        if typeCode is None:
            typeCode = self.typeConstant
        handler = self.handler.c_lookup( value )
        if isinstance( handler, FormatHandler ):
            return (<FormatHandler>handler).c_arraySize( 
                value, typeCode 
            )
        return handler.arraySize( 
            value, typeCode
        )
    def unitSize( self, value, typeCode=None ):
        """Determine unit size of an array (if possible)
        
        Uses our local type if defined, otherwise asks the handler to guess...
        """
        if typeCode is None:
            typeCode = self.typeConstant
        handler = self.handler.c_lookup( value )
        if isinstance( handler, FormatHandler ):
            return (<FormatHandler>handler).c_unitSize( 
                value, typeCode 
            )
        return handler.unitSize( 
            value, typeCode
        )
    def returnHandler( self ):
        """Get the default return-handler"""
        return self.handler.c_get_output_handler( )
    def zeros( self, dims, typeCode=None ):
        """Allocate a return array of the given dimensions filled with zeros"""
        return self.c_zeros( dims, typeCode )
    cdef c_zeros( self, dims, typeCode ):
        """C-level function to create empty array"""
        if typeCode is None:
            typeCode = self.typeConstant
        handler = self.handler.c_get_output_handler( )
        if isinstance( handler, FormatHandler ):
            return (<FormatHandler>handler).c_zeros( 
                dims, typeCode 
            )
        return handler.zeros( 
            dims, typeCode
        )
        
        
    def dimensions( self, value, typeCode=None ):
        """Given a data-value, get the dimensions (assumes full structure info)"""
        if typeCode is None:
            typeCode = self.typeConstant
        handler = self.handler.c_lookup( value )
        if isinstance( handler, FormatHandler ):
            return (<FormatHandler>handler).c_dimensions( 
                value
            )
        return handler.dimensions( value )
    
    def arrayByteCount( self, value, typeCode=None ):
        """Given a data-value, try to determine number of bytes it's final form occupies
        
        For most data-types this is arraySize() * atomic-unit-size
        """
        if typeCode is None:
            typeCode = self.typeConstant
        handler = self.handler.c_lookup( value )
        if isinstance( handler, FormatHandler ):
            return (<FormatHandler>handler).c_arrayByteCount( 
                value
            )
        return handler.arrayByteCount( value )

# Now some array helper functions...

cdef class Output(cArgConverter):
    """CConverter generating static-size typed output arrays
    
    Produces an output array of given type (arrayType) and 
    size using self.lookup() to determine the size of the 
    array to be produced, where the lookup function is passed 
    as an initialisation argument.
    
    Provides also:
    
        oldStyleReturn( ... ) for use in the default case of
            PyOpenGL compatability mode, where result arrays of
            size (1,) are returned as scalar values.
    """
    cdef str name 
    cdef tuple size
    cdef ArrayDatatype arrayType
    cdef int outIndex
    def __init__( self, str name, tuple size, ArrayDatatype arrayType ):
        self.name = name 
        self.size = size 
        self.arrayType = arrayType
    def finalise( self, wrapper ):
        self.outIndex = wrapper.cArgIndex( self.name )
        
    cdef tuple c_getSize( self, tuple pyArgs ):
        """Retrieve the array size for this argument"""
        return self.size
    
    cdef object c_call( self, tuple pyArgs, int index, object baseOperation ):
        """Return pyArgs[ self.index ]"""
        return self.arrayType.c_zeros( self.c_getSize(pyArgs), self.arrayType.typeConstant )
    
    def oldStyleReturn( self, object result, object baseOperation, tuple pyArgs, tuple cArgs ):
        """Retrieve cArgs[ self.index ]"""
        #TODO: make this a c_api-bearing value, not a Python function call
        cdef tuple thisSize
        result = cArgs[ self.outIndex ]
        try:
            thisSize = self.c_getSize(pyArgs)
        except KeyError as err:
            return result
        if thisSize == (1,):
            try:
                return result[0]
            except (IndexError,TypeError) as err:
                return result
        else:
            return result
cdef class OutputOrInput( Output ):
    cdef object c_call( self, tuple pyArgs, int index, object baseOperation ):
        """Return pyArgs[ self.index ]"""
        if pyArgs[index] is not DO_OUTPUT[0] and pyArgs[index] is not DO_OUTPUT[1]:
            return self.arrayType.asArray( pyArgs[index] )
        return self.arrayType.c_zeros( self.c_getSize(pyArgs), self.arrayType.typeConstant )

cdef class SizedOutput( Output ):
    """Output class that looks up output size via a callable function
    
    specifier -- Python argument name used to lookup the data-size 
    lookup -- function taking argument in specifier to determine size 
    """
    cdef str specifier
    cdef object lookup
    cdef int index 
    
    def __init__( self, str name, str specifier, object lookup, ArrayDatatype arrayType ):
        super( SizedOutput,self).__init__( name, None, arrayType )
        self.specifier = specifier
        self.lookup = lookup
        self.arrayType = arrayType
    def finalise( self, wrapper ):
        super( SizedOutput,self).finalise( wrapper )
        self.index = wrapper.pyArgIndex( self.specifier )
    cdef tuple c_getSize( self, tuple pyArgs ):
        """Retrieve the array size for this argument"""
        try:
            specifier = pyArgs[ self.index ]
        except AttributeError as err:
            raise RuntimeError( """"Did not resolve parameter index for %r"""%(self.name))
        else:
            try:
                result =  self.lookup( specifier )
                if not isinstance( result, tuple ):
                    result = (result,)
                return result
            except (TypeError,KeyError) as err:
                raise KeyError( """Unknown specifier %r (lookup = %s)"""%( specifier, self.lookup ))
DO_OUTPUT = (None,_NULL)
cdef class SizedOutputOrInput( SizedOutput ):
    cdef object c_call( self, tuple pyArgs, int index, object baseOperation ):
        """Return pyArgs[ self.index ]"""
        if pyArgs[index] is not DO_OUTPUT[0] and pyArgs[index] is not DO_OUTPUT[1]:
            return self.arrayType.asArray( pyArgs[index] )
        return self.arrayType.c_zeros( self.c_getSize(pyArgs), self.arrayType.typeConstant )
    

cdef class AsArrayOfType(pyArgConverter):
    """Given arrayName and typeName coerce arrayName to array of type typeName
    
    TODO: It should be possible to drop this if ERROR_ON_COPY,
    as array inputs always have to be the final objects in that 
    case.
    """
    cdef public str arrayName
    cdef public str typeName
    
    cdef int arrayIndex
    cdef int typeIndex
    cdef public ArrayDatatype arrayType
    
    def __init__( self, str arrayName='pointer', str typeName='type' ):
        self.arrayName = arrayName
        self.typeName = typeName 
        from OpenGL.arrays.arraydatatype import ArrayDatatype
        self.arrayType = ArrayDatatype
    def finalise( self, wrapper ):
        self.arrayIndex = wrapper.pyArgIndex( self.arrayName )
        self.typeIndex = wrapper.pyArgIndex( self.typeName )
    cdef object c_call( self, object incoming, object function, tuple arguments ):
        """Get the arg as an array of the appropriate type"""
        return self.arrayType.asArray( incoming, arguments[ self.typeIndex ] )

cdef class AsArrayTyped(pyArgConverter):
    """Given arrayName and arrayType, convert arrayName to array of type
    
    TODO: It should be possible to drop this if ERROR_ON_COPY,
    as array inputs always have to be the final objects in that 
    case.
    """
    cdef public str arrayName
    cdef int arrayIndex
    cdef public ArrayDatatype arrayType
    def __init__( self, arrayName='pointer', arrayType=None ):
        self.arrayName = arrayName # not actually used...
        if arrayType is None:
            from OpenGL.arrays.arraydatatype import ArrayDatatype
            arrayType = ArrayDatatype
        self.arrayType = arrayType
    def finalise( self, wrapper ):
        """Finalize the wrapper (nothing to do here)"""
    cdef object c_call( self, object incoming, object function, tuple arguments ):
        """Get the arg as an array of the appropriate type"""
        return self.arrayType.asArray( incoming )

cdef class AsArrayTypedSizeChecked( AsArrayTyped ):
    """Size-checking version of AsArrayTyped"""
    cdef int size 
    def __init__( self, arrayType=None, size=None ):
        super(AsArrayTypedSizeChecked,self).__init__( 'pointer', arrayType )
        baseSize = ctypes.sizeof( arrayType.baseType )
        self.size = size * baseSize
    cdef object c_call( self, object incoming, object function, tuple arguments ):
        """Get the arg as an array of the appropriate type"""
        cdef int actualSize
        result = self.arrayType.asArray( incoming )
        actualSize = self.arrayType.arrayByteCount( result )
        if actualSize != self.size:
            raise ValueError(
                """Expected %r byte array, got %r byte array"""%(
                    self.size,
                    actualSize,
                ),
                incoming,
            )
        return result
    
cdef class AsArrayTypedSize(cArgConverter):
    """Given arrayName and arrayType, determine size of arrayName
    """
    cdef public str arrayName
    cdef int arrayIndex
    cdef public ArrayDatatype arrayType
    
    def __init__( self, arrayName='pointer', arrayType=None ):
        self.arrayName = arrayName
        if arrayType is None:
            from OpenGL.arrays.arraydatatype import ArrayDatatype
            arrayType = ArrayDatatype
        self.arrayType = arrayType
    def finalise( self, wrapper ):
        self.arrayIndex = wrapper.pyArgIndex( self.arrayName )
    cdef c_call( self, tuple pyArgs, int index, object wrappedOperation ):
        return self.arrayType.arraySize( pyArgs[self.arrayIndex ] )

